import logging
import os

from msb.env import (NameConst, PathConst)
from django.db.migrations.recorder import MigrationRecorder

from .funcs import (log_to_console)


class DjangoBase:
	def _execute_cmd(self, *args, format=True, log=True):
		from django.core.management import call_command
		if log:
			log_to_console(msg=f"executing {' '.join(args if len(args) > 0 else 'all')} ", format=format)
		call_command(*args)

	def _db_query(self, db_connection, *queries):
		if db_connection and len(queries) > 0:
			with db_connection.cursor() as cursor:
				_result = []
				for query in queries:
					print(f"Executing Query = {query}\n")
					query_result = cursor.execute(query)
					if query.lstrip(' ').lower().startswith(('select', 'show')):
						_result.append(cursor.fetchall())
					else:
						_result.append(query_result)
				return _result[0] if len(_result) == 1 else _result


class DjangoFixtureInstaller(DjangoBase):

	def __init__(self, from_dir: str, app: str = '', is_yaml: bool = True):
		self.recorder = MigrationRecorder(None)
		self.fixture_dir = from_dir
		self.app_name = app
		self.file_ext = NameConst.YAML_FILE_EXTENTION_NAME if is_yaml else "json"

	@property
	def migration_objects(self):
		return self.recorder.Migration.objects

	@property
	def fixture_app(self):
		return f"{self.app_name + '_' if self.app_name  else ''}{self.fixture_dir}_fixtures".strip(' ')

	@property
	def fixture_directory_path(self):
		return PathConst.FIXTURES_DIR_PATH.joinpath(self.fixture_dir)

	@property
	def fixture_files_to_load(self) -> list:
		if os.path.isdir(self.fixture_directory_path):
			return sorted([file for file in os.listdir(self.fixture_directory_path) if file.endswith(f'.{self.file_ext}')])
		return []

	def _save_fixture_record(self, fixture):
		# Create a new migration record with the fixture name
		migration_record = self.recorder.Migration(app=self.fixture_app, name=fixture)

		# Save the migration record to the django_migrations table
		migration_record.save()

	def _get_installed_fixtures(self):
		# Retrieve all the installed fixtures from the django_migrations table
		installed_fixtures = self.migration_objects.filter(app=self.fixture_app).values_list('name', flat=True)

		return list(installed_fixtures)

	def _delete_fixture_record(self, *fixtures):
		# Delete the record for a particular fixture from the django_migrations table
		self.recorder.Migration.objects.filter(app=self.fixture_app, name__in=fixtures).delete()

	def _load_fixture(self, fixture):
		try:
			self._execute_cmd("loaddata", fixture, format=False, log=False)
			self._save_fixture_record(fixture=fixture)
			return "Success"
		except Exception as e:
			return f"Failed->{e}"

	def install(self, reinstall: bool = False):
		_log_message, _final_result = "Loading {} fixtures from '{}' Dir in following sequence:", []
		if (_fixture_count := len(_fixtures_list := self.fixture_files_to_load)) > 0:
			if reinstall:
				self._delete_fixture_record(*_fixtures_list)

			log_to_console(_log_message.format(_fixture_count, self.fixture_dir), format=True)
			_installed_fixtures = self._get_installed_fixtures()

			for _fixture in _fixtures_list:
				_status = self._load_fixture(_fixture) if _fixture not in _installed_fixtures \
					else "Skipped (Fixture is Already Installed)."
				print(_fixture, ":", _status)



from functools import wraps

from .utils import (validate_inp_payload, validate_inp_parameters, Request, ValidationSchema)


def validation_schema_wrapper(klass):
	def get(key: str, default=None) -> ValidationSchema | None:
		if hasattr(klass, key) and isinstance((schema := getattr(klass, key)), ValidationSchema):
			return schema
		return default

	setattr(klass, "get", get)
	return klass


def api_inputs(payload_schema: ValidationSchema = None, param_schema: ValidationSchema = None, **opt):
	allow_unknown = opt.get("unknown") == True
	bulk_inp = opt.get("bulk_inp") == True

	def outer_func(_func):
		@wraps(_func)
		def inner_func(cls, request: Request, *args, **kwargs):
			"""
			validate parameter data like from URL/Query string if parameter validation rules are defined
			"""
			if isinstance(param_schema, ValidationSchema):
				validate_inp_parameters(param_schema, kwargs, allow_unknown=True)

			"""
			validate payload data if payload rules are defined
			"""
			if isinstance(payload_schema, ValidationSchema):
				validate_inp_payload(payload_schema, request, allow_unknown, bulk_inp)

			cls._api_input_is_validated = True
			return _func(cls, *args, **dict(request=request, **kwargs))

		return inner_func

	return outer_func


def verify_search_employee_list(key: str = 'employee_number', parent_key: str = 'filters'):
	def decorator_function(func):
		@wraps(func)
		def wrapper_function(cls, request, *args, **kwargs):
			"""
			This function takes the key for employee number in search.
			It validates all the employee numbers in the payload if employee number is provided else will
			return the employee numbers of all the employee data the user has access to.
			"""
			# Get the employee number from the payload and make sure it is list
			employees = cls.payload[parent_key].get(key) or []
			employees_list = employees if type(employees) is list else [employees]
			# Validate and update the employee list
			cls.payload[parent_key].update({key: cls.user.get_validated_employee_list(*employees_list)})
			return func(cls, request, *args, **kwargs)

		return wrapper_function

	return decorator_function



from .decorators import (require_role, api_permissions, verify_ip)
from .permissions import (Permissions)
from .session import SessionData
from .users import TokenUser

__all__ = ['Permissions', 'TokenUser', 'require_role', 'api_permissions', 'verify_ip']



from django.db import models
from msb.cipher import Cipher

from .constants import (COLUMN_NAME_DELETED)


class MsbModelManager(models.Manager):
	_default_filters = {}
	_exclude_filters = {}

	def _get_filtered_queryset(self, **filters):
		_query_set = super(MsbModelManager, self).get_queryset()
		query_filters = {k: v for k, v in filters.items() if hasattr(self.model, k)}
		return _query_set.filter(**query_filters)

	def retrieve(self, *args, **kwargs):
		return super(MsbModelManager, self).get(*args, **kwargs)

	def get(self, *args, **kwargs):
		pk = kwargs.get('pk')
		if pk is not None and isinstance(pk, str):
			kwargs.update(dict(pk=int(Cipher.decipher(pk))))
		return self.retrieve(*args, **kwargs)

	@property
	def deleted(self):
		return self._get_filtered_queryset(**{COLUMN_NAME_DELETED: True})

	@property
	def from_all(self):
		return super(MsbModelManager, self).get_queryset()

	def get_queryset(self):
		return self._get_filtered_queryset(
			**{COLUMN_NAME_DELETED: False, **self._default_filters}
		).exclude(**self._exclude_filters)
